@@ -6,7 +6,10 @@ The Canvas object is the primary interface for creating PDF files. See
 doc/reportlab-userguide.pdf for copious examples.
 """
 
-__all__ = ['Canvas']
+__all__ = [
+        'Canvas',
+        'ShowBoundaryValue',
+        ]
 ENABLE_TRACKING = 1 # turn this off to do profile testing w/o tracking
 
 import re
@@ -16,10 +19,12 @@ from math import sin, cos, tan, pi
 from reportlab import rl_config
 from reportlab.pdfbase import pdfdoc
 from reportlab.pdfbase import pdfmetrics
+from reportlab.pdfbase.ttfonts import ShapedStr, shapeFragWord
 from reportlab.pdfgen  import pathobject
-from reportlab.pdfgen.textobject import PDFTextObject, _PDFColorSetter
+from reportlab.pdfgen.textobject import PDFTextObject, _PDFColorSetter, bidiShapedText
 from reportlab.lib.colors import black, _chooseEnforceColorSpace, Color, CMYKColor, toColor
 from reportlab.lib.utils import ImageReader, isSeq, isStr, isUnicode, _digester, asUnicode
+from reportlab.lib.abag import ABag
 from reportlab.lib.rl_accel import fp_str, escapePDF
 from reportlab.lib.boxstuff import aspectRatioFix
 
@@ -47,7 +52,7 @@ PATH_OPS = {(0, 0, FILL_EVEN_ODD) : 'n',  #no op
             }
 
 def _annFormat(D,color,thickness,dashArray,hradius=0,vradius=0):
-    from reportlab.pdfbase.pdfdoc import PDFArray, PDFDictionary
+    from reportlab.pdfbase.pdfdoc import PDFArray
     if color and 'C' not in D:
         D["C"] = PDFArray([color.red, color.green, color.blue])
     if 'Border' not in D:
@@ -209,6 +214,15 @@ def _gradientExtendStr(extend):
         return "[%s %s]" % ['true' if _ else 'false' for _ in extend]
     return "[true true]" if extend else "[false false]"
 
+class ShowBoundaryValue:
+    def __init__(self,color=(0,0,0),width=0.1,dashArray=None):
+        self.color = color
+        self.width = width
+        self.dashArray = dashArray
+
+    def __bool__(self):
+        return self.color is not None and self.width>=0
+
 class Canvas(_PDFColorSetter):
     """This class is the programmer's interface to the PDF file format.  Methods
     are (or will be) provided here to do just about everything PDF can do.
@@ -277,6 +291,7 @@ class Canvas(_PDFColorSetter):
                  trimBox=None,
                  bleedBox=None,
                  lang=None,
+                 **kwds,
                  ):
         """Create a canvas of a given size. etc.
 
@@ -354,6 +369,7 @@ class Canvas(_PDFColorSetter):
         self.state_stack = []
 
         self.setEncrypt(encrypt)
+        self._namedCB = {}  #named callbacks
 
     def setEncrypt(self, encrypt):
         '''
@@ -1116,9 +1132,9 @@ class Canvas(_PDFColorSetter):
         """
         #check if we've done this one already...
         if isUnicode(command):
-            rawName = 'PS' + hashlib.md5(command.encode('utf-8')).hexdigest()
+            rawName = 'PS' + hashlib.md5(command.encode('utf-8'),usedforsecurity=False).hexdigest()
         else:
-            rawName = 'PS' + hashlib.md5(command).hexdigest()
+            rawName = 'PS' + hashlib.md5(command,usedforsecurity=False).hexdigest()
         regName = self._doc.getXObjectName(rawName)
         psObj = self._doc.idToObject.get(regName, None)
         if not psObj:
@@ -1594,9 +1610,9 @@ class Canvas(_PDFColorSetter):
         # use PDFTextObject for multi-line text.
         ##################################################
 
-    def drawString(self, x, y, text, mode=None, charSpace=0, direction=None, wordSpace=None):
+    def drawString(self, x, y, text, mode=None, charSpace=0, direction=None, wordSpace=None, shaping=False):
         """Draws a string in the current text styles."""
-        text = asUnicode(text)
+        text, width = bidiShapedText(text,direction,fontName=self._fontname,fontSize=self._fontsize, shaping=shaping)
         #we could inline this for speed if needed
         t = self.beginText(x, y, direction=direction)
         if mode is not None: t.setTextRenderMode(mode)
@@ -1608,14 +1624,12 @@ class Canvas(_PDFColorSetter):
         if mode is not None: t.setTextRenderMode(0)
         self.drawText(t)
 
-    def drawRightString(self, x, y, text, mode=None, charSpace=0, direction=None, wordSpace=None):
+    def drawRightString(self, x, y, text, mode=None, charSpace=0, direction=None, wordSpace=None, shaping=False):
         """Draws a string right-aligned with the x coordinate"""
-        if not isinstance(text, str):
-            text = text.decode('utf-8')
-        width = self.stringWidth(text, self._fontname, self._fontsize)
+        text, width = bidiShapedText(text,direction,fontName=self._fontname,fontSize=self._fontsize, shaping=shaping)
         if charSpace: width += (len(text)-1)*charSpace
         if wordSpace: width += (text.count(u' ')+text.count(u'\xa0')-1)*wordSpace
-        t = self.beginText(x - width, y, direction=direction)
+        t = self.beginText(x - width, y)
         if mode is not None: t.setTextRenderMode(mode)
         if charSpace: t.setCharSpace(charSpace)
         if wordSpace: t.setWordSpace(wordSpace)
@@ -1625,16 +1639,14 @@ class Canvas(_PDFColorSetter):
         if mode is not None: t.setTextRenderMode(0)
         self.drawText(t)
 
-    def drawCentredString(self, x, y, text, mode=None, charSpace=0, direction=None, wordSpace=None):
+    def drawCentredString(self, x, y, text, mode=None, charSpace=0, direction=None, wordSpace=None, shaping=False):
         """Draws a string centred on the x coordinate. 
         
         We're British, dammit, and proud of our spelling!"""
-        if not isinstance(text, str):
-            text = text.decode('utf-8')
-        width = self.stringWidth(text, self._fontname, self._fontsize)
+        text, width = bidiShapedText(text,direction,fontName=self._fontname,fontSize=self._fontsize, shaping=shaping)
         if charSpace: width += (len(text)-1)*charSpace
         if wordSpace: width += (text.count(u' ')+text.count(u'\xa0')-1)*wordSpace
-        t = self.beginText(x - 0.5*width, y, direction=direction)
+        t = self.beginText(x - 0.5*width, y)
         if mode is not None: t.setTextRenderMode(mode)
         if charSpace: t.setCharSpace(charSpace)
         if wordSpace: t.setWordSpace(wordSpace)
@@ -1644,7 +1656,7 @@ class Canvas(_PDFColorSetter):
         if mode is not None: t.setTextRenderMode(0)
         self.drawText(t)
 
-    def drawAlignedString(self, x, y, text, pivotChar=rl_config.decimalSymbol, mode=None, charSpace=0, direction=None, wordSpace=None):
+    def drawAlignedString(self, x, y, text, pivotChar=rl_config.decimalSymbol, mode=None, charSpace=0, direction=None, wordSpace=None, shaping=False):
         """Draws a string aligned on the first '.' (or other pivot character).
 
         The centre position of the pivot character will be used as x.
@@ -1686,19 +1698,19 @@ class Canvas(_PDFColorSetter):
                 leftText = leftText[0:-1]
 
             self.drawRightString(x-0.5*pivW, y, leftText, mode=mode, charSpace=charSpace,
-                    direction=direction, wordSpace=wordSpace)
+                    direction=direction, wordSpace=wordSpace, shaping=shaping)
             self.drawString(x-0.5*pivW, y, rightText, mode=mode, charSpace=charSpace,
-                    direction=direction, wordSpace=wordSpace)
+                    direction=direction, wordSpace=wordSpace, shaping=shaping)
 
         else:
             #normal case
             leftText = parts[0]
             self.drawRightString(x-0.5*pivW, y, leftText, mode=mode, charSpace=charSpace,
-                    direction=direction, wordSpace=wordSpace)
+                    direction=direction, wordSpace=wordSpace, shaping=shaping)
             if len(parts) > 1:
                 rightText = pivotChar + parts[1]
                 self.drawString(x-0.5*pivW, y, rightText, mode=mode, charSpace=charSpace,
-                        direction=direction, wordSpace=wordSpace)
+                        direction=direction, wordSpace=wordSpace, shaping=shaping)
 
     def getAvailableFonts(self):
         """Returns the list of PostScript font names available.
@@ -1977,12 +1989,33 @@ class Canvas(_PDFColorSetter):
             self._doc._catalog.AcroForm = self.AcroForm = AcroForm(self)
             return self.AcroForm
 
-    @property
-    def drawBoundary(self):
-        if not hasattr(self,'_drawBoundary'):
-            from reportlab.platypus import Frame
-            self._drawBoundary = lambda sb,x,y,w,h: Frame._drawBoundary(self,sb,x,y,w,h)
-        return self._drawBoundary
+    def drawBoundary(self,sb,x1,y1,width,height):
+        "draw a boundary as a rectangle (primarily for debugging)."
+        ss = isinstance(sb,(str,tuple,list)) or isinstance(sb,Color)
+        w = -1
+        da = None
+        if ss:
+            c = toColor(sb,-1)
+            ss = c != -1
+        elif isinstance(sb,ShowBoundaryValue) and sb:
+            c = toColor(sb.color,-1)
+            ss = c != -1
+            if ss:
+                w = sb.width
+                da = sb.dashArray
+        if ss:
+            self.saveState()
+            self.setStrokeColor(c)
+            if w>=0: self.setLineWidth(w)
+            if da: self.setDash(da)
+        self.rect(x1,y1,width,height)
+        if ss: self.restoreState()
+
+    def setNamedCB(self, name, cb):
+        self._namedCB[name] = cb
+
+    def getNamedCB(self,name):
+        return self._namedCB.get(name,None)
 
 if __name__ == '__main__':
     print('For test scripts, look in tests')
